/*
 * Title:        bits/visitors.hpp
 * Description:  C++ header-only wrapper for C-ABI data_t type
 * $Copyright (c) 2015-2016 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: visitors.hpp 320837 2017-11-30 21:07:30Z bsp $
 */

/* This file:
 * Contains the C++ implementation of visitor and const_visitor
 *
 */

#ifndef _DATAT_BITS_INCLUDE_TAG
#error Must only be included from sag_connectivity_cpp.hpp, use that instead
#endif


/**
 * Helper class for writing visitors to apply to data_t.
 *
 * This class provides the result_type typedef and all the operator()() methods and
 * should be used for visiting non-const data_ts. For visiting const data_ts
 * then you need to use const_visitor
 *
 * These then call to visitTYPE methods which default to an error condition.
 *
 * Users should override the visitTYPE(TYPE&) const methods for the types which their visitor needs to handle.
 * Unhandled types will still default to error.
 *
 * The default error condition is to throw std::runtime_error. 
 * If you wish to have a different error condition then override the error() method.
 *
 * This class is templated over both the derived class (to avoid the need for virtual calls) and the return type.
 * It should be used as:
 *
 * class my_visitor: public visitor<my_visitor, void>
 * {
 * ...
 * };
 */
template<typename DERIVED, typename RV>
class visitor
{
public:
	/** The return type from applying this visitor */
	typedef RV result_type;
	/** The type of the derived class */
	typedef DERIVED derived_t;

	/** Visit an empty variant (don't override this, instead override visitEmpty) */
	result_type operator()() const { return derived().visitEmpty(); }
	/** Visit an empty variant */
	result_type visitEmpty() const { return derived().error("empty"); }

	/** Visit an integer (don't override this, instead override visitInteger) */
	result_type operator()(int64_t &i) const { return derived().visitInteger(i); }
	/** Visit an integer */
	result_type visitInteger(int64_t &i) const { return derived().error("integer"); }

	/** Visit a double (don't override this, instead override visitDouble) */
	result_type operator()(double &i) const { return derived().visitDouble(i); }
	/** Visit a double */
	result_type visitDouble(double &i) const { return derived().error("double"); }

	/** Visit a boolean (don't override this, instead override visitBoolean) */
	result_type operator()(bool &i) const { return derived().visitBoolean(i); }
	/** Visit a boolean */
	result_type visitBoolean(bool &i) const { return derived().error("boolean"); }

	/** Visit a string (don't override this, instead override visitString) */
	result_type operator()(const char *&i) const { return derived().visitString(i); }
	/** Visit a string */
	result_type visitString(const char *&i) const { return derived().error("string"); }

	/** Visit a decimal (don't override this, instead override visitDecimal) */
	result_type operator()(decimal_t &i) const { return derived().visitDecimal(i); }
	/** Visit a decimal */
	result_type visitDecimal(decimal_t &i) const { return derived().error("decimal"); }

	/** Visit a byte buffer (don't override this, instead override visitBuffer) */
	result_type operator()(buffer_t &i) const { return derived().visitBuffer(i); }
	/** Visit a byte buffer */
	result_type visitBuffer(buffer_t &i) const { return derived().error("buffer"); }

	/** Visit a list (don't override this, instead override visitList) */
	result_type operator()(list_t &i) const { return derived().visitList(i); }
	/** Visit a list */
	result_type visitList(list_t &i) const { return derived().error("list"); }

	/** Visit a map (don't override this, instead override visitMap) */
	result_type operator()(map_t &i) const { return derived().visitMap(i); }
	/** Visit a map */
	result_type visitMap(map_t &i) const { return derived().error("map"); }

	/** Visit a custom object (don't override this, instead override visitCustom) */
	result_type operator()(sag_underlying_custom_t &i) const { return derived().visitCustom(i); }
	/** Visit a custom object */
	result_type visitCustom(sag_underlying_custom_t &i) const { return derived().error("custom"); }

protected:
	/** Handles visiting unhandled types.
		Defaults to throwing std::runtime_error. Override it to provide different defaults. */
	result_type error(const std::string &reason) const { throw std::runtime_error("Visit type error: "+reason); }
private:
	/** \internal Cast this to the derived type to call overridden methods */
	const derived_t &derived() const { return static_cast<const derived_t&>(*this); }
};

/**
 * Helper class for writing visitors to apply to data_t.
 *
 * This class provides the result_type typedef and all the operator()() methods and
 * should be used for visiting const data_ts. or non-const data_ts that you
 * don't require non-const access to. If you need to modify the data_t contents then
 * use visitor.
 *
 * These then call to visitTYPE methods which default to an error condition.
 *
 * Users should override the visitTYPE(const TYPE&) const methods for the types which their visitor needs to handle.
 * Unhandled types will still default to error.
 *
 * The default error condition is to throw std::runtime_error. 
 * If you wish to have a different error condition then override the error() method.
 *
 * This class is templated over both the derived class (to avoid the need for virtual calls) and the return type.
 * It should be used as:
 *
 * class my_visitor: public visitor<my_visitor, void>
 * {
 * ...
 * };
 */
template<typename DERIVED, typename RV>
class const_visitor
{
public:
	/** The return type from applying this visitor */
	typedef RV result_type;
	/** The type of the derived class */
	typedef DERIVED derived_t;

	/** Visit an empty variant (don't override this, instead override visitEmpty) */
	result_type operator()() const { return derived().visitEmpty(); }
	/** Visit an empty variant */
	result_type visitEmpty() const { return derived().error("empty"); }

	/** Visit an integer (don't override this, instead override visitInteger) */
	result_type operator()(int64_t i) const { return derived().visitInteger(i); }
	/** Visit an integer */
	result_type visitInteger(int64_t i) const { return derived().error("integer"); }

	/** Visit a double (don't override this, instead override visitDouble) */
	result_type operator()(double i) const { return derived().visitDouble(i); }
	/** Visit a double */
	result_type visitDouble(double i) const { return derived().error("double"); }

	/** Visit a boolean (don't override this, instead override visitBoolean) */
	result_type operator()(bool i) const { return derived().visitBoolean(i); }
	/** Visit a boolean */
	result_type visitBoolean(bool i) const { return derived().error("boolean"); }

	/** Visit a string (don't override this, instead override visitString) */
	result_type operator()(const char *i) const { return derived().visitString(i); }
	/** Visit a string */
	result_type visitString(const char *i) const { return derived().error("string"); }

	/** Visit a decimal (don't override this, instead override visitDecimal) */
	result_type operator()(const decimal_t &i) const { return derived().visitDecimal(i); }
	/** Visit a decimal */
	result_type visitDecimal(const decimal_t &i) const { return derived().error("decimal"); }

	/** Visit a byte buffer (don't override this, instead override visitBuffer) */
	result_type operator()(const buffer_t &i) const { return derived().visitBuffer(i); }
	/** Visit a byte buffer */
	result_type visitBuffer(const buffer_t &i) const { return derived().error("buffer"); }

	/** Visit a list (don't override this, instead override visitList) */
	result_type operator()(const list_t &i) const { return derived().visitList(i); }
	/** Visit a list */
	result_type visitList(const list_t &i) const { return derived().error("list"); }

	/** Visit a map (don't override this, instead override visitMap) */
	result_type operator()(const map_t &i) const { return derived().visitMap(i); }
	/** Visit a map */
	result_type visitMap(const map_t &i) const { return derived().error("map"); }

	/** Visit a custom object (don't override this, instead override visitCustom) */
	result_type operator()(const sag_underlying_custom_t &i) const { return derived().visitCustom(i); }
	/** Visit a custom object */
	result_type visitCustom(const sag_underlying_custom_t &i) const { return derived().error("custom"); }

protected:
	/** Handles visiting unhandled types.
		Defaults to throwing std::runtime_error. Override it to provide different defaults. */
	result_type error(const std::string &reason) const { throw std::runtime_error("Visit type error: "+reason); }
private:
	/** \internal Cast this to the derived type to call overridden methods */
	const derived_t &derived() const { return static_cast<const derived_t&>(*this); }
};

