#!/usr/bin/env python
# Copyright (c) 2015 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors. 
# Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG 

import glob, os
from pysys.constants import PROJECT, TIMEOUTS, FOREGROUND, BACKGROUND

class UMRealmHelper:
	"""Helper class for the Universal Messaging Realm server. Attempts to detect and flag errors from the log/stderr wherever possible.

	Non-returning methods are Fluent - that is, they return 'self' to allow method-chaining
	"""
	def __init__(self, parent, port, serverout, host = "localhost"):
		""" Create an instance of UMRealmHelper for manipulating an existing realm server on an NSP protocol. Only returns when the server
		is up and ready for commands to be run against it.

		@param host: The hostname of the UM realm server
		@param port: The port (NSP) of the UM realm server
		@param parent: the PySys testcase using this
		@param serverout: Filename prefix for realm server output; a suffix of '.out' is the logfile, a suffix of '.err' is stderr.
		"""
		self.parent = parent
		self.port = port
		self.host = host
		self.serverout = serverout
		matches=self.parent.waitForSignal(file=serverout+'.out', expr='Realm Server Startup sequence completed', timeout=TIMEOUTS['WaitForSignal']*4)
		if len(matches) != 1:
			raise Exception("Realm server has not started successfully (see %s.out/err)" % serverout)
		self.__checkLogs()
		self.parent.addResource(self)

	def command(self, jclass, args=None, abortOnError=True, state=FOREGROUND):
		""" Invoke a UM client utility against this realm server

		@param jclass: The Java class behind the utility (we didn't want to bother parsing the named Java wrapper scripts)
		@param args: List of arguments to the utility
		@param abortOnError: If the command fails, throw an exception if true
		@param state: Run the process either in the C{FOREGROUND} or C{BACKGROUND} (defaults to C{FOREGROUND})
		"""
  
		# NUM-13145 .../lib/umSampleApplications.jar (the replacement) for .../classes - we add all jars so covered in next line
		classpath = ":".join(glob.glob(PROJECT.UM_SAG_HOME + "/UniversalMessaging/lib/*.jar")) + ":" + (PROJECT.UM_SAG_HOME + "/UniversalMessaging/classes/")
		displayName = "UM-" + jclass[jclass.rfind(".") + 1:]
		output = displayName + "_" + str(self.parent.getInstanceCount(displayName))
		rname = "nsp://" + self.host + ":" + str(self.port)
		run=self.parent.startProcess(command=PROJECT.JAVA_HOME + "/bin/java",
		                             arguments=["-DRNAME=" + rname, "-cp", classpath, jclass] + (args if args else []),
		                             displayName=displayName, stderr=output + ".err", stdout=output + ".out", workingDir=self.parent.output,
		                             timeout=TIMEOUTS['WaitForSignal'],
		                             state=state)
		if(run.exitStatus != 0 and abortOnError and state==FOREGROUND):
			raise Exception("execution of " + jclass + " failed")

		return self

	def __checkLogs(self):
		""" Check for failures in the UM server output, called during startup and after commands in order to be very robust in our testing """
		# This grep specifically avoids the entirely-expected stack traces at shut down, by virtue of different indentation!
		self.parent.assertGrep(file=self.serverout + '.out', expr='^\tat com\.', contains=False)
		self.parent.assertGrep(file=self.serverout + '.err', expr='at com\.', contains=False)
	
	def __del__(self):
		self.__checkLogs()
